#!/usr/bin/perl -w
# Like CladeAgeGenes.pl but uses BLAST thresholds only
#

use strict;
use lib "$ENV{HOME}/Genomics/Perl/modules";
use GenomicsUtils;
use Gene;
use GeneMethods;
use Genome;
use Getopt::Long;

{
    my $usage = "CladeAgeGenes.pl [-clades cladeLists -ignore cladeList | -genomes genomesFile]\n"
	. "  -out prefix [-analyze 1]\n"
	. "  -threshold 1.3 -host lut [-debug 0]\n"
	.  "   The rule for HGT is: s(n) > threshold * max(s(n-1), s(n-2))\n"
	.  "       where s(n) is the best BLAST score within a clade (or 0).\n"
	.  "   The rule for ORFan is existence of a BLAST hit (which is not perfect)\n"
	.  "   The rule for Native is present (by BLAST) in all clades and not HGT\n"
	.  "   Reports genomes in prefix.out, gene categories in prefix.genes, potential\n"
	.  "   sources for HGT events in prefix.hits, and quartets to test to validate HGT in\n"
	.  "   prefix.quartets. Any hit outside of the in-group with a score above best/threshold\n"
	.  "   is included in prefix.hits\n"
	.  "   If using a genomesFile, it should have the taxId in the second column and the clade in the 3rd;\n"
	.  "   genomes not in the file are ignored\n";

    my $genomeFile = undef;
    my $cladesSpec = undef;
    my $ignore = "";
    my $out = undef;
    my $host = undef;
    my $debug = 0;
    my $threshold = 1.3;
    my $analyze=1;

    (GetOptions('clades=s' => \$cladesSpec,
		'ignore=s' => \$ignore,
		'genomes=s' => \$genomeFile,
		'host=s' => \$host,
		'debug=i' => \$debug,
		'threshold=f' => \$threshold,
		'analyze=i' => \$analyze,
		'out=s' => \$out)
     && @ARGV==0 && (defined $cladesSpec xor defined $genomeFile)
     && defined $out && defined $host) || die $usage;

    setDebug($debug);
    connect($host);

    my @ignore = ();
    my %taxToIgnore = ();
    my @alltax = (); # not including ignored set
    my @clades = (); # list of list of taxonomy ids
    my %taxToClade = ();

    if (defined $genomeFile) {
	die "Cannot specify ignore with genomeFile" if defined $ignore && $ignore ne "";
	open(GENOMES,"<",$genomeFile) || die "Cannot read $genomeFile";
	while(<GENOMES>) {
	    my ($value, $taxId, $clade) = split /\t/, $_;
	    next if $clade eq "Clade";
	    die "Bad line $_" unless defined $clade && $taxId =~ m/^[0-9-]+$/ && $clade =~ m/^[0-9]+$/;
	    die "TaxId $taxId is present more than once" if exists $taxToClade{$taxId};
	    $taxToClade{$taxId} = $clade;
	    push @{ $clades[$clade] }, $taxId; 
	    push @alltax, $taxId;
	    my $taxValid = GenomicsUtils::queryScalar("SELECT count(distinct scaffoldId) FROM Scaffold"
						      . " WHERE taxonomyId=$taxId AND isGenomic=1"
						      . " AND isActive=1");
	    print STDERR "Warning: Obsolete or invalid taxonomyId $taxId (clade $clade)\n" if $taxValid==0;
	}
	close(GENOMES) || die "Error reading $genomeFile";
	@ignore = grep {!exists $taxToClade{$_}} getChildTaxa(131567);
	%taxToIgnore = map { $_ => 1 } @ignore;
    } else {
	@ignore = getChildTaxa(split /,/, $ignore);
	@alltax = grep { !exists $taxToIgnore{$_} } getChildTaxa(131567); # cellular organisms (excluding viruses etc.)
	%taxToIgnore = map { $_ => 1 } @ignore;
	my @cladeSpecs = split /:/, $cladesSpec;
	die $usage unless @cladeSpecs > 1;

	foreach my $cladeSpec (@cladeSpecs) {
	    my @show = grep { !exists $taxToIgnore{$_}; } getChildTaxa(split /,/, $cladeSpec);
	    my @new = grep { !exists $taxToClade{$_}; } @show;
	    die "Invalid clade specification $cladeSpec -- no new non-ignored genomes in it" unless @new;
	    foreach (@new) { $taxToClade{$_} = scalar @clades; }
	    push @clades, \@new;
	}
    }
    my @taxOut = grep { !exists $taxToClade{$_} } @alltax;
    my %taxOut = map { $_ => "" } @taxOut;

    open(GENOMES, ">", "$out.genomes") || die "Cannot write to $out.genomes";
    print GENOMES join("\t","N","Tax","Clade","Name")."\n";
    my $n = 0;
    foreach my $index (0..$#clades) {
	my @new = @{ $clades[$index] };
	my %tNames = ();
	foreach my $tax (@new) {
	    $tNames{$tax} = Genome::new('taxonomyId'=>$tax)->shortName();
	}
	foreach my $tax (sort { $tNames{$a} cmp $tNames{$b} } @new) {
	    print GENOMES join("\t", ++$n, $tax, $index, $tNames{$tax}) . "\n";
	}
    }
    my %tNames = map { $_ => Genome::new('taxonomyId'=>$_)->shortName() } @taxOut;
    foreach my $tax (sort {$tNames{$a} cmp $tNames{$b}} @taxOut) {
	print GENOMES join("\t",++$n, $tax, $#clades+1, $tNames{$tax})."\n";
    }
    close(GENOMES) || die "Cannot close $out.genomes";
    print STDERR "Wrote to $out.genomes\n";
    exit(0) unless $analyze;


    # Then load locus->taxonomyId mapping
    print STDERR "Loading locusId->taxonomyId mapping\n";
    my $results = &query("SELECT s.taxonomyId, l.locusId FROM Locus l, Scaffold s"
			 . " WHERE l.priority=1 AND l.scaffoldId=s.scaffoldId AND s.isActive=1");
    my %locusToTax = ();
    my %taxToClade0 = ();
    my @targets = ();
    foreach (@{ $clades[0] }) { $taxToClade0{$_} = 1; }
    foreach (@$results) {
	my ($tax, $locus) = @$_;
	$locusToTax{$locus} = $tax;
	if (exists $taxToClade0{$tax}) { push @targets, $locus; }
    }

    $| = 1;
    open(GENES, ">", "$out.genes") || die;
    # BestN is the best blast score for clade N
    # MinSAge is the lowest possible source age (clade) (for HGT genes only)
    # SAge is the source age with the best hit
    my @names = map { "Best" . $_ } (0..($#clades+1));
    print GENES join("\t", "locusId", "Type", "Age", "MinSAge", "SAge",@names)."\n";
    open(HITS, ">", "$out.hits") || die;
    # only potential sources for HGT
    # N is position within the list of hits for this gene (1 is best hit)
    print HITS join("\t", "locusId","Hit","N","Clade","Tax","Score")."\n";
    open(QUARTETS,">","$out.quartets") || die;
    print QUARTETS join("\t","locusId","Out","Intermediate1","Intermediate2")."\n";
    print STDERR "Searching for BLAST hits\n";
    foreach my $target (@targets) {
	my $results = &query("SELECT subject,score FROM `BLASTp_".$locusToTax{$target}."`"
			     . " WHERE locusId=$target");
	next if @$results == 0; # cannot age if no blast hits, e.g. for RNA genes, pseudogenes, or short ORFs
	my %hits = (); # gene -> score
	foreach (@$results) {
	    my ($subject,$score) = @$_;
	    $hits{$subject} = $score unless exists $hits{$subject} && $hits{$subject} > $score;
	}
	my %best = (); # clade -> [best hit, best score]
	my ($HIT,$SCORE) = (0,1);
	foreach (0..($#clades+1)) { $best{$_} = [-1,0]; }
	foreach my $subject (keys %hits) {
	    my $tax = $locusToTax{$subject};
	    next unless defined $tax && !exists $taxToIgnore{$tax};
	    my $clade =  $taxToClade{$tax};
	    $clade = $#clades+1 if !defined($clade) && exists $taxOut{$tax};
	    next unless defined $clade;
	    $best{$clade} = [$subject,$hits{$subject}] unless $best{$clade}[$SCORE] > $hits{$subject};
	}
	my $age = undef;
	my $type = "";
	my $minsource = "";
	my $bestsource = "";
	$type = "Native";
	if ($best{1}[1] > 0) {
	    $type = "Native";
	    $age = 1;
	} else {
	    $type = "ORFan";
	    $age = 0;
	}
	
	foreach my $clade (2..($#clades+1)) {
	    if ($best{$clade}[$SCORE] > $threshold*$best{$clade-1}[$SCORE]
		&& $best{$clade}[$SCORE] > $threshold*$best{$clade-2}[$SCORE]) {
		$age = $clade-3;
		while ($age > 0 && $best{$age}[$SCORE] < $best{$clade}[$SCORE]/$threshold) { $age--; }
		$type = "HGT";
		$minsource = $clade;
		$bestsource = $clade;
		foreach my $clade2 (($clade+1)..($#clades+1)) {
		    $bestsource = $clade2 if $best{$clade2}[$SCORE] > $best{$bestsource}[$SCORE];
		}
		last;
	    } elsif ($best{$clade}[$SCORE] == 0) {
		$age = $clade-1 unless $type eq "ORFan";
		$type = "ORFan";
	    } elsif ($best{$clade}[$SCORE] > 0) {
		$age = $clade;
		$type = "" if $type eq "ORFan"; # unclear
	    }
	}
	my @scores = map $best{$_}[$SCORE], (0..($#clades+1));
	print GENES join("\t", $target, $type, $age, $minsource, $bestsource, @scores) . "\n";
	if ($type eq "HGT") {
	    my $n = 0;
	    my $min = $best{$bestsource}[$SCORE] / $threshold;
	    foreach my $subject (sort {$hits{$b} <=> $hits{$a}} (keys %hits)) {
		if ($hits{$subject} > $min) {
		    my $tax = $locusToTax{$subject};
		    next unless defined $tax && !exists $taxToIgnore{$tax};
		    my $clade =  $taxToClade{$tax};
		    $clade = $#clades+1 if !defined($clade) && exists $taxOut{$tax};
		    next if !defined $clade || $clade <= $age;
		    print HITS join("\t", $target, $subject, ++$n, $clade, $tax, $hits{$subject})."\n";
		}
	    }
	    my @intermediate = ();
	    foreach my $clade (($age+1)..($minsource-1)) {
		push @intermediate, $best{$clade}[$HIT] if ($best{$clade}[$SCORE] > 0);
		last if @intermediate == 2;
	    }
	    print QUARTETS join("\t", $target, $best{$bestsource}[$HIT], @intermediate)."\n" if @intermediate==2;
	}
    }
    close(GENES);
    close(HITS);
    close(QUARTETS);
    print STDERR "Done\n";
    disconnect();
}
